#include "Matrioska\MatrioskaHeader.hpp"



Mat currentFrame, frame;
Mat currentFrameForHist;

Rect objectROI;
int frameCount;

bool isObjectFound = false;

Mat objectModel;


float firstRange[] = { 0, 180 };				// Hue
float secondRange[] = { 0, 256 };				// Saturation
const float* histRange[2];						//= { firstRange, secondRange };

int channels[] = { 0, 1 };
int nChannels = 2;

int firstBins = 80;
int secondBins = 90;
int histBins[] = { firstBins, secondBins };

void normalizeHistogram(Mat& hist, int pixels = 0){

	normalize(hist, hist, 255, 0, NORM_L1);

	//hist /= pixels;
	//hist *= 255;
}

void updateModel(){

	Mat currentObject = currentFrameForHist(objectROI);

	// Get the Histogram and normalize it
	calcHist(&currentObject, 1, channels, Mat(), 
		objectModel, nChannels, histBins, histRange, true, false);

	//normalize(objectModel, objectModel, 0, 255, NORM_MINMAX, -1, Mat());
	//normalize(objectModel, objectModel, 255, 0, NORM_L2, -1, Mat());
	

	normalizeHistogram(objectModel, currentObject.rows*currentObject.cols);


	cout << "Object histogram: " << objectModel.size() <<", depth: "<<objectModel.depth()<<", channels: "<<objectModel.channels()<< endl;
	//cout << objectModel << endl;
	cout << "Sum value: " << sum(objectModel) << endl;
	
	

	showImage(objectModel, "Object Model");
}


void updateModelWithBackground(){

	Rect backgroundRect = rectScale(objectROI, 3.0f);
	rectInsideImage(currentFrameForHist, backgroundRect);


	Mat detectionMask = Mat(currentFrameForHist.size(), CV_8U, Scalar::all(0));
	
	detectionMask(backgroundRect).setTo(Scalar::all(255));
	detectionMask(objectROI).setTo(Scalar::all(0));

	int pixels = (detectionMask.rows * detectionMask.cols) - (objectROI.width * objectROI.height);

	/*
	Mat bg;showImage(detectionMask,"background mask");currentFrameForHist.copyTo(bg, detectionMask);
	showImage(bg,"masked");waitKey(0);*/
	
	Mat backgroundModel;

	// Get the Histogram and normalize it
	calcHist(&currentFrameForHist, 1, channels, detectionMask,
		backgroundModel, nChannels, histBins, histRange, true, false);

	normalizeHistogram(backgroundModel, pixels);
	showImage(backgroundModel, "Background model");

	cout << "Background model sum: " << sum(backgroundModel) << endl;

	// default 20.0f;
	float scaleFactor = 20.0f;

	// Binarization
	for (int i = 0; i < objectModel.rows; i++) {
		for (int j = 0; j < objectModel.cols; j++) {

			float obj = objectModel.at<float>(i, j);
			float bg = backgroundModel.at<float>(i, j);

			if (obj>bg*scaleFactor)
				objectModel.at<float>(i, j) = 255;
			else
				objectModel.at<float>(i, j) = 0;
			
			//cout << "Background value: " << (int)bg << endl;
			//cout << "Object value: " << (int)objectModel.at<uchar>(i, j) << endl;


			
		}
	}


	showImage(objectModel, "object model filtered");
	
}

class colorBinarization{



public:


};

float colorTracking(string datasetName = "") {



	// === 
	// DATASET
	// ===


	GTManager myGT;
	VideoCapture cap;
	VideoMaker video;

	//video.setWritePath("G:/LTDT/_outputMatrioska/video/");
	//myGT.setOutputFile("G:/LTDT/" + datasetName + "/Matrioska.txt");


	string queryPath = "F:/Value/Personal/Parthenope/Project/input/babenko/VOTChallenge/" + datasetName + "/%08d.jpg";
	string init = "F:/Value/Personal/Parthenope/Project/input/babenko/VOTChallenge/" + datasetName + "/groundtruth.txt";


	/*
	string queryPath = "G:/LTDT/" + datasetName + "/%05d.jpg";
	string init = "G:/LTDT/" + datasetName + "/gt_mkv.txt";
	myGT.useTLBR = true;
	*/

	cap.open(queryPath);
	myGT.setGroundTruth(init);


	Counter counters;
	frameCount = 0;

	while (1) {


		Timer timerFull;

		cap >> currentFrame;

		//> Check if there are other frames to process (with webcam you need to manually exit with ESC key)
		if (currentFrame.empty())
			break;

		Rect gt = myGT.getNextRect();
		frameCount++;
		if (frameCount == 1) {
			objectROI = gt;
		}


		//cvtColor(currentFrame, currentFrame, COLOR_BGR2GRAY);

		currentFrame.copyTo(frame);
		if (currentFrame.channels() > 1) {
			cvtColor(currentFrame, currentFrameForHist, COLOR_BGR2HSV);	// Histogram
			cvtColor(currentFrame, currentFrame, COLOR_BGR2GRAY);

			histRange[0] = firstRange;
			histRange[1] = secondRange;

		} else {
			// if the dataset is grayscale
			currentFrameForHist = currentFrame;

			nChannels = 1;	
			firstRange[1] = 256;		// hue can be max 180, grayscale is 256 (exclusive)
			histRange[0] = firstRange;
			

			cvtColor(frame, frame, CV_GRAY2BGR);						// Frame should be colored for drawing functions
		}

		

		Log() << endl << endl << "Parsing frame #" << frameCount << endl;


		// Init
		if (frameCount == 1) {

			
			//objectROI = Rect(281, 111, 14, 17);
			cout << "Object ROI: " << objectROI << endl;

			

			
			
			
			updateModel();
			updateModelWithBackground();

			isObjectFound = true;

			drawRect(frame, objectROI);
			showImage(frame);
			waitKey(1000);

			myGT.calcOverlap(objectROI);

			continue;
		}

		//resize(currentFrameForHist, currentFrameForHist, Size(), 0.5, 0.5);

		// Get Backprojection
		Mat backproj;
		calcBackProject(&currentFrameForHist, 1, channels, objectModel, backproj, histRange);

		//equalizeHist(backproj, backproj);


		showImage(backproj, "Back Projected");


		/*
		if (votes>goodConfidenceVotes)
			isObjectFound = true;
		else
			isObjectFound = false;
		*/


		timerFull.timerEnd("Frame");


		if (isObjectFound) {


			drawRect(frame,objectROI,Scalar(255,0,0));
		}
		else
			objectROI = Rect(0, 0, 0, 0);


		//GT
		//drawRect(frame,myGT.boundingBoxGT,Scalar(255,255,255));

		//showImage(frame);

		if (!keyboardManagerStandard(500)) {
			break;
		}


		float overlap = myGT.calcOverlap(objectROI);
		Log() << "Overlap Accuracy: " << overlap << endl;

		//video.writeFrame(frame,true);

	} // while loop

	myGT.printStats();


	destroyAllWindows();
	return 0.0f;
}


int main(){



	//colorTracking("iceskater");


	vector<string> dataset;
	dataset.push_back("bicycle");
	dataset.push_back("bolt");
	dataset.push_back("car");
	dataset.push_back("cup");
	dataset.push_back("david");
	dataset.push_back("diving");
	dataset.push_back("face");
	dataset.push_back("gymnastics");
	dataset.push_back("hand");
	dataset.push_back("iceskater");
	dataset.push_back("juice");
	dataset.push_back("jump");
	dataset.push_back("singer");
	dataset.push_back("sunshade");
	dataset.push_back("torus");
	dataset.push_back("woman");

	dataset.clear();
	//dataset.push_back("jump");dataset.push_back("torus");dataset.push_back("woman");
	dataset.push_back("diving");

	vector<float> datasetAccuracy;
	for (auto& name : dataset) {
		datasetAccuracy.push_back(colorTracking(name));
		waitKey(100);
	}



	sp();
	return 0;


}