/*
Copyright  2013 Mario Edoardo Maresca
 
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:
 
The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.
 
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


using namespace std;
using namespace cv;

//> Helper functions

	


	//> Pause the system
	void sp() {
		cout << endl << "Press Return to continue..."<<endl;
		cin.get();
	}

	//> Convert myType i to string
	template <class myType>
	string toString(myType i) {
		stringstream ss;
		string s;
		ss << i;
		s = ss.str();
		return s;
	}

	template<class T>
	void printVector(vector<T>& v) {

		cout << "Vector size: " << v.size() << endl;
		for(auto i=0;i<v.size();i++)
			cout <<i<<" => ["<<v.at(i)<<"]" << endl;
	}

	template<class MyMap>
	void printMap(MyMap& m) {

		cout << "Map size: " << m.size() << endl;
		for( auto it = m.begin(); it != m.end(); ++it ) {
			cout <<"["<<it->first<<"] => ["<<it->second<<"]"<<endl;
		}

	}

	template<class MyMap>
	void printMultiMap(MyMap& m) {
		for(auto& i : m)
			for(auto& j : m[i.first])
				cout << "["<<i.first<<"]["<<j.first<<"]=>["<<m[i.first][j.first]<<"]"<<endl;
	}

	//> Show an image, the user can resize it
	void showImage(const Mat& image,string name = "Image") {
		namedWindow(name,CV_WINDOW_NORMAL);
		imshow(name,image);
	}

	//> Load an image from file
	void loadImage(string path, Mat& image, int flags = CV_LOAD_IMAGE_GRAYSCALE ) {

		cout << "Reading ["+(path.substr(path.rfind('/')-5))+"] " << endl;

		image = imread(path,flags);
		if (image.empty()) {
			cout << "Image not loaded ["<<path<<"]"<<endl;
			sp();
		}
	}



	//> Calc the mean of a vector of Point
	template <class myType>
	myType meanPoint(vector<myType> points) {

		//> Summation
		myType sum = accumulate(
			points.begin(), points.end(),	// Run from begin to end
			myType(0,0)						// Initialize with a zero point
		);
	
		//> Divide by count to get mean
		myType mean;

		if (!points.empty()) {
			mean.x = sum.x / (int)points.size();
			mean.y = sum.y / (int)points.size();
		}

		return mean;
	}

	// Calc a mean of elements inside a vector
	template<class myType>
	float meanVector(vector<myType> elements) {
		float sum = accumulate(elements.begin(), elements.end(), 0.0f) / elements.size();
		return sum;
	}

	//> Round to nearest. We are using a template because you can pass as argument both float or double
	template <class myType>
	int roundToNearest(myType num) {
		return (num > 0.0) ? (int)floor(num + 0.5) : (int)ceil(num - 0.5);
	}


	//> Calc distance between two points
	float euclideanDist(const Point2f& p, const Point2f& q) {
		Point2f diff = p - q;
		return (float)sqrt(diff.x*diff.x + diff.y*diff.y);
	}

	template<class myType>
	myType medianVector(std::vector<myType> vec) {
		if(vec.empty()) return 0;
		else {
			std::sort(vec.begin(), vec.end());
			return vec[vec.size() / 2];
			/*
			if(vec.size() % 2 == 0)
				return (vec[vec.size()/2 - 1] + vec[vec.size()/2]) / 2;
			else
				return vec[vec.size()/2];
			*/
		}
	}

	float randFloat(float a, float b){
		return a + static_cast<float>(rand()) / (static_cast<float>(RAND_MAX/(b-a)));
	}


	//> ================
	//> ROI/Rect related
	//> ================

	//> The rect r must be inside the image boundaries
	void rectInsideImage(const Mat& img, Rect& r) {

		Point topLeft = r.tl();
		Point bottomRight = r.br();

		//> Check Boundaries
		if (topLeft.x<0) topLeft.x=0;
		if (topLeft.y<0) topLeft.y=0;
		if (topLeft.x>img.cols) topLeft.x = img.cols;
		if (topLeft.y>img.rows) topLeft.y = img.rows; 

		if (bottomRight.x<0) bottomRight.x=0;
		if (bottomRight.y<0) bottomRight.y=0;
		if (bottomRight.x>img.cols) bottomRight.x = img.cols;
		if (bottomRight.y>img.rows) bottomRight.y = img.rows; 

		r = Rect(topLeft,bottomRight);
	}

	Rect getRectInsideImage(const Mat& img, const Rect& r) {
		Rect newRect = r;
		rectInsideImage(img, newRect);
		return newRect;
	}

	//> The rect r must be inside the image boundaries
	bool rectInsideImageCheck(const Mat& img, const Rect& r) {

		Point topLeft = r.tl();
		Point bottomRight = r.br();

		//> Check Boundaries
		if (topLeft.x<0 ||
			topLeft.y<0 ||
			topLeft.x>img.cols ||
			topLeft.y>img.rows ||
			bottomRight.x<0 ||
			bottomRight.y<0 ||
			bottomRight.x>img.cols ||
			bottomRight.y>img.rows)
		{
			return false;
		}

		return true;
		
	}

	//> Enlarges or shrinks a Rect, keeping costant the center
	Rect rectSum(Rect roi, Size pixels) {

		Point topLeft = roi.tl();
		Point bottomRight = roi.br();

		Point p = Point(pixels.width/2,pixels.height/2);
		topLeft -= p;
		bottomRight += p;

		return Rect(topLeft,bottomRight);
	}
	Rect rectSum(Rect roi, int pixels) {
		return rectSum(roi,Size(pixels,pixels));
	}
	Rect rectSum2(Rect& roi, int pixels) {
		return Rect(
			Point2i(roi.tl().x - pixels, roi.tl().y - pixels),
			Point2i(roi.br().x + pixels, roi.br().y + pixels)
		);

	}
	
	//> Height and Width must be greater than minSide 
	bool rectSizeCheck(const Rect& r, int minSide) {
		return (r.height>=minSide && r.width>=minSide);
	}

	//> Given a RotatedRect calc the Rect (a smaller version than boundingRect)
	Rect rectFromRotatedRect(const RotatedRect& rRoi) {

		auto W = rRoi.size.width;
		auto H = rRoi.size.height;

		Point2f center = rRoi.center;
		Point2f topLeft( rRoi.center.x - W/2,rRoi.center.y - H/2);
		Point2f bottomRight( rRoi.center.x + W/2,rRoi.center.y + H/2);

		return Rect(topLeft,bottomRight);

		//> We could simply use .boundingRect but it's better to have less background captured
		//return rRoi.boundingRect();
	}


	// Moves the rect r according to displacement d
	void rectUpdate(Rect& r, const Point& d) {
		r = Rect( r.tl() + d, r.br() + d );
	}


	// Keeps the same center and scales width and height
	Rect rectScale(const Rect& r, float scale) {
		Point center = Point( r.x + r.width/2, r.y + r.height/2);

		int width = roundToNearest(  r.width*scale ),
			height = roundToNearest( r.height*scale );

		Point tl(center.x-width/2, center.y-height/2);

		return Rect( tl.x, tl.y, width, height);
	}

	Rect rectScale(const Rect& r, Size2f scale) {
		Point center = Point(r.x + r.width / 2, r.y + r.height / 2);

		int width = roundToNearest( r.width*scale.width ),
			height = roundToNearest( r.height*scale.height );

		Point tl(center.x - width / 2, center.y - height / 2);

		return Rect(tl.x, tl.y, width, height);
	}

	Rect rectPerturb(const Rect& r) {
		
		Rect scaled = rectScale(r, randFloat(0.55f,0.95f));

		int shift = (int)randFloat(-5.0f,9.0f);
		return Rect(scaled.tl().x + shift,scaled.tl().y + shift,scaled.width,scaled.height);
	}

	Point2f rectCenter(const Rect& r) {

		return Point2f(r.tl().x + r.width / 2, r.tl().y + r.height / 2);

	}

	vector<Point2f> rectToCorners(const Rect& r){
		vector<Point2f> corners;
		corners.push_back(Point2f(r.x, r.y));
		corners.push_back(Point2f(r.br().x, r.br().y));
		corners.push_back(Point2f(r.x + r.width, r.y));
		corners.push_back(Point2f(r.x, r.y + r.height));
		return corners;
	}

	// Check if a point is inside an image
	bool pointInsideImage(const Mat& img, const Point& p) {
		return ( p.x > 0 && p.y > 0 && p.x<img.cols && p.y<img.rows );
	}



	Rect getNextSearchWindow(const Rect& win, int searchWindowSize, int windowOverlap = 30) {

		static int searchWindowId = 0;

		int winSize = (searchWindowSize-windowOverlap);
		int maxX = (win.width / winSize) + ((win.width%winSize)==0?0:1);
		int maxY = (win.height / winSize) + ((win.height%winSize)==0?0:1);
		
		Rect nextWindow( 
			((int)(searchWindowId%maxX)) * winSize, 
			((int)(searchWindowId/maxX)) * winSize, 
			searchWindowSize, searchWindowSize
		 );

		if ( (nextWindow.x + nextWindow.width) > win.width)
			nextWindow.width = win.width - nextWindow.x;

		if ( (nextWindow.y + nextWindow.height) > win.height)
			nextWindow.height = win.height - nextWindow.y;

		nextWindow.x += win.x;
		nextWindow.y += win.y;

		searchWindowId++;
		if ( searchWindowId+1 > (maxX*maxY) )
			searchWindowId = 0;

		return nextWindow;
	}


	
	Rect getNextSearchWindowRand(const Mat& image, int searchWindowSize) {

		// Cut some border pixels
		int cut = 30;

		int maxX = max(1,(image.cols-cut*2) - searchWindowSize);
		int maxY = max(1,(image.rows-cut*2) - searchWindowSize);
		
		Rect nextWin = Rect( 
			cut + (rand()%maxX), 
			cut + (rand()%maxY), 
			searchWindowSize, searchWindowSize
		);

		if ( nextWin.x+nextWin.width > image.cols )
			nextWin.width = (image.cols - nextWin.x) - cut;

		if ( nextWin.y+nextWin.height > image.rows )
			nextWin.height = (image.rows - nextWin.y) - cut;

		return nextWin;
	}

	//> ============= END ROI/Rect related








	//> ================
	//> DRAWING FUNCTION
	//> ================

	//> Wrapper for circle
	void drawCircle(Mat &img, const Point& center, Scalar color = Scalar(0,255,0))  {
		circle(img,center,1,color,1);
	}

	void drawPoints(Mat &img, const vector<Point>& points, Scalar color = Scalar(0,255,0)) {
		for(auto& p : points)
			drawCircle(img, p, color);
	}

	// Draw a filled circle
	void drawPoint(Mat &img, const Point& center, Scalar color = Scalar(0,255,0))  {
		drawCircle(img,center,color);
	}

	//> Draw a KeyPoint
	void drawSingleKP(Mat& img, const KeyPoint& kp){

		vector<KeyPoint> KPs;
		KPs.push_back(kp);
		drawKeypoints( img, KPs,img,Scalar::all(-1),4);

	}
	// Warp cv::drawKeypoints call ()
	void drawKeypoints(Mat& img, const vector<KeyPoint>& KPs, Scalar color = Scalar(0,255,0), bool small = false) {
		if (!small) {
			drawKeypoints( img, KPs, img, color, 4);
		} else {
			
			for(const auto& kp : KPs) {
				circle(img,kp.pt,1,color,1);
			}

		}
	}
	
	//> Having a point
	void drawSingleKP(Mat& img, const Point2f& p, float angle, float size=10){
		KeyPoint kp(p,size,angle,1,1);
		drawSingleKP( img, kp);
	}

	//> Mark a point with a red circle
	void drawMark(Mat& img, const Point& p, Scalar color = Scalar(0,0,255)) {
		circle(img, p, 2, color, CV_FILLED );
	}
	void drawMark(Mat& img, const KeyPoint& kp, Scalar color = Scalar(0, 0, 255)) {
		drawMark(img,kp.pt,color);
	}

	void drawMarkX(Mat& img, const KeyPoint& kp){
		float length = 3.0f;
		Scalar color(0,0,255);

		// Diagonal line
		Point2f line1pt1(kp.pt.x-length,kp.pt.y-length);
		Point2f line1pt2(kp.pt.x+length,kp.pt.y+length);
		line(img, line1pt1,line1pt2, color, 1);

		// Anti diagonal
		Point2f line2pt1(kp.pt.x+length,kp.pt.y-length);
		Point2f line2pt2(kp.pt.x-length,kp.pt.y+length);
		line(img, line2pt1,line2pt2, color, 1);

	}

	//> Draw a bounding box
	void drawRect(Mat& im, const Rect& roi, Scalar color = Scalar(0,255,0), int thickness = 1) {
		rectangle(im,roi.tl(),roi.br(),color,thickness);
	}

	void drawRect(Mat& frame,const vector<Point2f>& vertices, Scalar color = Scalar(0,255,0)){
		for (auto i = 0; i < 4; i++)
			line(frame, vertices[i], vertices[(i+1)%4], color);
	}

	void drawRectCenter(Mat& im, const Rect& roi, Scalar color = Scalar(0,255,0)) {
		Point2i c( roi.tl().x+(roi.width/2), roi.tl().y+(roi.height/2) );
		drawPoint(im,c,color);
		
	}
	
	void drawRect(Mat& frame,Point2f* vertices, Scalar color = Scalar(0,255,0)){
		for (int i = 0; i < 4; i++)
			line(frame, vertices[i], vertices[(i+1)%4], Scalar(0,255,0));
	}
	void drawRect(Mat& frame, const RotatedRect& rRect,Scalar color = Scalar(0,255,0)) {
		Point2f vertices[4];
		rRect.points(vertices);
		drawRect(frame,vertices,color);
	}

	void drawCross(Mat& im, const Rect& roi) {
		
		Point2i center( roi.tl().x + roi.width/2, roi.tl().y + roi.height/2 );

		//> Vertical Line
		line(im,Point((int)center.x,0),Point((int)center.x,im.size().height),(0,0,0),2);
	
		//> Horizontal
		line(im,Point(0,(int)center.y),Point(im.size().width,(int)center.y),(0,0,0),2);

		circle(im,center,20,Scalar(0,0,0),1);
	
	}





	//> Draw a grid of uniform size inside the image
	//> With distance = 0, the distance will be infered from gridRows and gridCols
	//> If distance != 0, gridRows and gridCols will be ignored
	void drawGrid(Mat& im, int gridRows, int gridCols, int dist = 0) {

		int width=im.size().width;
		int height=im.size().height;

		Scalar color(0,0,0);
		int distH;
		int distW;
		if (dist==0) {
			distH = height/gridRows;
			distW = width/gridCols;
		} else {
			distH = distW = dist;
		}


		//> Draw the grid
		for(int i=0;i<height;i+=distH)
			line(im,Point(0,i),Point(width,i),color);

		for(int i=0;i<width;i+=distW)
			line(im,Point(i,0),Point(i,height),color);

	}
	void drawGrid(Mat& im, int dist) {
		drawGrid(im,1,1,dist);
	}

	// Draw a line between two keypoints
	void drawLine(Mat& im, const KeyPoint& kp1, const KeyPoint& kp2) {
		line(im,kp1.pt,kp2.pt,Scalar(0,255,0));
	}

	//> =============== END DRAWING FUNCTION

	
	
	//> This methods selects just one keypoints (with the greatest response) inside a kernel of size SidexSide
	void filterKeyPointsByKernel(vector<KeyPoint>& kp, int side = 2) {

		unordered_map<int,unordered_map<int,vector<KeyPoint>>> gridKP;

		//> Create the reference map with buckets
		for(auto i=0;i<kp.size();i++) {
			KeyPoint& k = kp[i];

			int x = (int)k.pt.x / side;
			int y = (int)k.pt.y / side;

			gridKP[x][y].push_back(k);
		}

		//> Selects the best keypoints inside each bucket
		vector<KeyPoint> filteredKP;
		for(auto& x: gridKP) {
			for(auto& y: gridKP[x.first]) {

				float maxResponse = 0;
				size_t maxID = 0;

				//> Bucket traversing
				for(auto j = 0;j<gridKP[x.first][y.first].size();j++) {

					KeyPoint& k = gridKP[x.first][y.first][j];

					if ( k.response > maxResponse ) {
						//cout << " Current response: "<<maxResponse << ", this response: "<<k.response<<endl;
						maxResponse = k.response;
						maxID = j;
					}

				}

				filteredKP.push_back( gridKP[x.first][y.first][maxID] );
			}
		}

		//cout << "Filtered: " + toString(kp.size() - filteredKP.size()) + " keypoints, out of " + toString( kp.size() )
		//		+ " total keypoints. Filtered size: "+toString(filteredKP.size());

		kp.swap(filteredKP);



	}

	// Rotate an image
	void rotateImage(Mat& source, Point2f& center, double angle, Mat& dst) {
			Mat M = getRotationMatrix2D(center, angle, 1.0);
			warpAffine(source, dst, M, source.size());
	}


	// Shift all keypoints with a given x and y (shift) (needed when using the ROI)
	void shiftKeyPoints(vector<KeyPoint>& kp, const Point2f& shift) {

		// If there is no shift returns
		if (shift.x==0 && shift.y==0)
			return;

		// For each point of keypoints add the shift
		for(auto i=0;i<kp.size();i++) {
			kp[i].pt += shift;
		}
	}

	// Manage pause and esc functions
	bool keyboardManagerStandard(int delay = 1) {

		char c = waitKey(delay);
		if (c == 27) {
			exit(0);
			return false;
		} else if (c == 'p')
			sp();
	
		return true;

	}

	// Given two set of points, calc the pairwise distance and estimate the scale with the median
	float scaleEstimation(const vector<Point>& previous, const vector<Point>& current){

		assert(previous.size() == current.size());
			
		// Too few points
		if (previous.size() < 3)
			return 1.0f;

		vector<float> scales;

		for (auto i = 0; i < previous.size()/2; ++i){

			float previousDistance = euclideanDist(previous[i], previous[i*2]);
			if (previousDistance < 3.0f)
				continue;

			float currentDistance = euclideanDist(current[i], current[i*2]);
	
			float scale = currentDistance / previousDistance;
			

			scales.push_back(scale);
		}
		

		float scale = medianVector<float>(scales);
		return scale;

	}

	// Non-isotropic scale factor
	Size2f scaleEstimation2(const vector<Point>& previous, const vector<Point>& current){

		assert(previous.size() == current.size());

		// Too few points
		if (previous.size() < 4)
			return Size2f(1.0f,1.0f);

		vector<float> scalesX;
		vector<float> scalesY;

		for (int i = 0; i < (int)(previous.size() / 2); ++i){

			float previousDistanceX = previous[i].x - previous[i * 2].x;
			if (previousDistanceX < 2.0f)
				continue;

			float currentDistanceX = current[i].x - current[i * 2].x;
			float scaleX = currentDistanceX / previousDistanceX;
			if (scaleX<0.4f || scaleX>1.4f)
				continue;
			scalesX.push_back(scaleX);

			float previousDistanceY = previous[i].y - previous[i * 2].y;
			if (previousDistanceY < 2.0f)
				continue;
			float currentDistanceY = current[i].y - current[i * 2].y;
			float scaleY = currentDistanceY / previousDistanceY;
			if (scaleY<0.7f || scaleY>1.3f)
				continue;
			scalesY.push_back(scaleY);


		}

		if (scalesX.empty())
			scalesX.push_back(1.0f);
		if (scalesY.empty())
			scalesY.push_back(1.0f);


		float scaleX = meanVector<float>(scalesX);
		float scaleY = meanVector<float>(scalesY);

		return Size2f(scaleX,scaleY);

	}


	// Calculate the atan2 and return the angle in degrees. The result is mapped to [0,360)
	float degreeAtan2(float y, float x) {

		float angle = atan2(y, x) * radToDegree;
		if (angle < 0)
			angle += 360;
		//float angle = atan(y / x) * radToDegree;
		return angle;

	}

	// Check if a file exists
	bool fileExists(const std::string& name) {
		ifstream f(name.c_str());
		return f.good();
	}

	void fileAppend(const std::string& name, const std::string& content) {

		std::ofstream outfile;

		outfile.open(name, std::ios_base::app);
		outfile << content;
	}


	// External class

	template<typename kind>
	class CircularBuffer {
	private:
		kind *buffer;
		int write_element;
		int read_element;
		
	public:
		size_t buffSize;
		CircularBuffer(size_t buff) : write_element(0), read_element(0) {
			buffSize = buff;
			buffer = new kind[buffSize];
		}

		~CircularBuffer() {
			delete[] buffer;
		}

		void append(kind value) {
			if (write_element >= buffSize) {
				write_element = 0;
			}

			buffer[write_element] = value;

			++write_element;
			read_element = write_element + 1;
		}

		kind get() {
			if (read_element >= buffSize) {
				read_element = 0;
			}

			++read_element;
			return(buffer[(read_element - 1)]);
		}

		int current() {
			return read_element;
		}
	};